package de.hft.stuttgart.citydoctor2.webservice.endpoints;

import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.time.LocalDateTime;

import javax.persistence.EntityManager;
import javax.servlet.http.HttpServletRequest;
import javax.ws.rs.Consumes;
import javax.ws.rs.FormParam;
import javax.ws.rs.GET;
import javax.ws.rs.POST;
import javax.ws.rs.Path;
import javax.ws.rs.Produces;
import javax.ws.rs.core.Context;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;
import javax.ws.rs.core.Response.Status;

import org.apache.logging.log4j.Level;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.webservice.AuthenticationException;
import de.hft.stuttgart.citydoctor2.webservice.database.TokenQueries;
import de.hft.stuttgart.citydoctor2.webservice.database.UserQueries;
import de.hft.stuttgart.citydoctor2.webservice.model.LoginSuccessfulReturnType;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.Token;
import de.hft.stuttgart.citydoctor2.webservice.model.entities.User;
import de.hft.stuttgart.citydoctor2.webservice.utils.AuthenticationUtils;
import de.hft.stuttgart.citydoctor2.webservice.utils.HashUtils;
import de.hft.stuttgart.citydoctor2.webservice.utils.HexUtils;
import de.hft.stuttgart.citydoctor2.webservice.utils.HibernateUtils;

@Path("/login")
public class LoginEndpoint {

	private static Logger logger = LogManager.getLogger(LoginEndpoint.class);

	@POST
	@Produces(MediaType.APPLICATION_JSON)
	@Consumes(MediaType.APPLICATION_FORM_URLENCODED)
	public Response authenticateUser(@FormParam("username") String username, @FormParam("password") String password,
			@Context HttpServletRequest request) {
		logger.trace("Login attempt from {}, password hidden", username);
		if (username == null || password == null) {
			return Response.status(Status.BAD_REQUEST).build();
		}
		EntityManager manager = HibernateUtils.createManager();
		try {
			User user = UserQueries.getUserForUsername(username, manager);
			if (user == null) {
				throw new AuthenticationException("Unknown user");
			}
			authenticate(user, password);
			String addr = request.getLocalAddr();
			Token token = TokenQueries.getTokenForUser(user, addr, manager);
			if (token == null) {
				token = createToken(user, addr);
				TokenQueries.persistToken(token, manager);
			} else {
				LocalDateTime expires = token.getExpires();
				if (expires.isBefore(LocalDateTime.now())) {
					manager.getTransaction().begin();
					manager.remove(token);
					manager.getTransaction().commit();
					// token expired, issue new one
					token = createToken(user, addr);
					TokenQueries.persistToken(token, manager);
				}
			}
			logger.trace("Login successful by user {}", user.getId());
			LoginSuccessfulReturnType returnType = new LoginSuccessfulReturnType(token.getTokenString(),
					token.getExpires(), user.getId());
			return Response.status(Status.OK).entity(returnType).build();
		} catch (AuthenticationException e) {
			logger.trace("Authentication denied for user {}, reason: {}", username, e);
			return Response.status(Status.FORBIDDEN).build();
		} catch (Exception e) {
			logger.catching(Level.FATAL, e);
			return Response.status(Status.INTERNAL_SERVER_ERROR).build();
		} finally {
			manager.close();
		}
	}
	
	@GET
	public Response logout(@Context HttpServletRequest request) {
		EntityManager manager = null;
		try {
			manager = HibernateUtils.createManager();
			AuthenticationUtils.removeTokenFromUser(request, manager);
			return Response.ok().build();
		} catch (AuthenticationException e) {
			logger.trace(AuthenticationUtils.AUTHENTICATION_DENIED, request.getRemoteAddr(), e);
			return Response.status(Status.FORBIDDEN).build();
		} finally {
			if (manager != null) {
				manager.close();
			}
		}
	}

	private Token createToken(User user, String addr) {
		byte[] bytes = new byte[16];
		SecureRandom r = new SecureRandom();
		r.nextBytes(bytes);
		String tokenString = HexUtils.encode(bytes);

		Token token = new Token();
		token.setAddress(addr);
		token.setUser(user);
		token.setExpires(LocalDateTime.now().plusHours(3));
		token.setTokenString(tokenString);
		return token;
	}

	private void authenticate(User user, String password) throws AuthenticationException, NoSuchAlgorithmException {
		String hash = HashUtils.createPasswordHash(password, user.getSalt());
		String storedPassword = user.getPasswordHash();
		if (!storedPassword.equals(hash)) {
			throw new AuthenticationException("Wrong password");
		}
	}
}
